// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0.If a copy of the MPL was not distributed with this
// file, You can obtain one at https ://mozilla.org/MPL/2.0/.

#ifndef SPHERICAL_H
#define SPHERICAL_H

#include "Direction.h"

namespace Passer {
namespace LinearAlgebra {

struct Vector3;
template <typename T> class PolarOf;

template <typename T> class SphericalOf {
public:
  /// @brief The distance in meters
  /// @remark The distance should never be negative
  float distance;
  /// @brief The angle in the horizontal plane in degrees, clockwise rotation
  /// @details The angle is automatically normalized to -180 .. 180
  // AngleOf<T> horizontal;
  /// @brief The angle in the vertical plane in degrees. Positive is upward.
  /// @details The angle is automatically normalized to -180 .. 180
  // AngleOf<T> vertical;
  DirectionOf<T> direction;

  SphericalOf<T>();
  SphericalOf<T>(float distance, AngleOf<T> horizontal, AngleOf<T> vertical);
  SphericalOf<T>(float distance, DirectionOf<T> direction);

  static SphericalOf<T> FromPolar(PolarOf<T> v);

  static SphericalOf<T> FromVector3(Vector3 v);
  Vector3 ToVector3() const;

  /// @brief A spherical vector with zero degree angles and distance
  const static SphericalOf<T> zero;
  /// @brief A normalized forward-oriented vector
  const static SphericalOf<T> forward;
  /// @brief A normalized back-oriented vector
  const static SphericalOf<T> back;
  /// @brief A normalized right-oriented vector
  const static SphericalOf<T> right;
  /// @brief A normalized left-oriented vector
  const static SphericalOf<T> left;
  /// @brief A normalized up-oriented vector
  const static SphericalOf<T> up;
  /// @brief A normalized down-oriented vector
  const static SphericalOf<T> down;

  SphericalOf<T> WithDistance(float distance);

  /// @brief Negate the vector
  /// @return The negated vector
  /// This will rotate the vector by 180 degrees horizontally and
  /// vertically. Distance will stay the same.
  SphericalOf<T> operator-() const;

  /// @brief Subtract a spherical vector from this vector
  /// @param v The vector to subtract
  /// @return The result of the subtraction
  SphericalOf<T> operator-(const SphericalOf<T> &v) const;
  SphericalOf<T> operator-=(const SphericalOf<T> &v);
  /// @brief Add a spherical vector to this vector
  /// @param v The vector to add
  /// @return The result of the addition
  SphericalOf<T> operator+(const SphericalOf<T> &v) const;
  SphericalOf<T> operator+=(const SphericalOf<T> &v);

  /// @brief Scale the vector uniformly up
  /// @param f The scaling factor
  /// @return The scaled vector
  /// @remark This operation will scale the distance of the vector. The angle
  /// will be unaffected.
  friend SphericalOf<T> operator*(const SphericalOf<T> &v, float f) {
    return SphericalOf<T>(v.distance * f, v.direction);
  }
  friend SphericalOf<T> operator*(float f, const SphericalOf<T> &v) {
    return SphericalOf<T>(f * v.distance, v.direction);
  }
  SphericalOf<T> operator*=(float f);
  /// @brief Scale the vector uniformly down
  /// @param f The scaling factor
  /// @return The scaled factor
  /// @remark This operation will scale the distance of the vector. The angle
  /// will be unaffected.
  friend SphericalOf<T> operator/(const SphericalOf<T> &v, float f) {
    return SphericalOf<T>(v.distance / f, v.direction);
  }
  friend SphericalOf<T> operator/(float f, const SphericalOf<T> &v) {
    return SphericalOf<T>(f / v.distance, v.direction);
  }
  SphericalOf<T> operator/=(float f);

  static float DistanceBetween(const SphericalOf<T> &v1,
                               const SphericalOf<T> &v2);
  static AngleOf<T> AngleBetween(const SphericalOf<T> &v1,
                                 const SphericalOf<T> &v2);
  static AngleOf<T> SignedAngleBetween(const SphericalOf<T> &v1,
                                       const SphericalOf<T> &v2,
                                       const SphericalOf<T> &axis);

  static SphericalOf<T> Rotate(const SphericalOf &v, AngleOf<T> horizontalAngle,
                               AngleOf<T> verticalAngle);
  static SphericalOf<T> RotateHorizontal(const SphericalOf<T> &v,
                                         AngleOf<T> angle);
  static SphericalOf<T> RotateVertical(const SphericalOf<T> &v,
                                       AngleOf<T> angle);
};

using SphericalSingle = SphericalOf<float>;
using Spherical16 = SphericalOf<signed short>;

} // namespace LinearAlgebra
} // namespace Passer
using namespace Passer::LinearAlgebra;

#include "Polar.h"
#include "Vector3.h"

#endif